from __future__ import annotations

import json
import os
from pathlib import Path, PurePath
from shutil import copytree, ignore_patterns
from tempfile import gettempdir
from typing import Any, Dict, List, Union

import click
from loguru import logger
from pydantic import BaseSettings, Field, PydanticValueError, validator
from typing_extensions import Literal

from .utils import trash

__novalidate__ = '__novalidate__'


class _PathNotWritableDirectoryError(PydanticValueError):
    code = 'path.not_writable_directory'
    msg_template = '"{path}" is not a writable directory'


def _get_default_config_dir() -> Path:
    return Path(click.get_app_dir('instawow'))


def _validate_expand_path(value: Path) -> Path:
    try:
        return value.expanduser().resolve()
    except RuntimeError as error:
        # pathlib will raise a ``RuntimeError`` for non-existent ~users
        raise ValueError(str(error)) from error


def _validate_path_is_writable_dir(value: Path) -> Path:
    if not (value.is_dir() and os.access(value, os.W_OK)):
        raise _PathNotWritableDirectoryError(path=value)

    return value


class BaseConfig(BaseSettings):
    def _build_values(
        self,
        init_kwargs: Dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        # Prioritise env vars
        return {**init_kwargs, **self._build_environ()}


class GlobalConfig(BaseConfig):
    config_dir: Path = Field(default_factory=_get_default_config_dir)
    profile: str = Field('__default__', min_length=1, strip_whitespace=True)
    addon_dir: Path
    game_flavour: Literal['retail', 'classic']
    auto_update_check: bool = True
    temp_dir: Path = Path(gettempdir()) / 'instawow'

    class Config:  # type: ignore
        env_prefix = 'INSTAWOW_'

    @validator('config_dir', 'addon_dir', 'temp_dir')
    def _validate_expand_path(cls, value: Path) -> Path:
        return _validate_expand_path(value)

    @validator('addon_dir')
    def _validate_path_is_writable_dir(cls, value: Path) -> Path:
        if value.name == __novalidate__:
            return value
        return _validate_path_is_writable_dir(value)

    @staticmethod
    def is_classic_folder(folder: str) -> bool:
        tail = PurePath(folder).parts[-3:]
        return tuple(map(str.casefold, tail)) == ('_classic_', 'interface', 'addons')

    @classmethod
    def get_dummy_config(cls, **kwargs: Any) -> GlobalConfig:
        "Create a dummy configuration with default values."
        template = {'game_flavour': 'retail', 'addon_dir': __novalidate__}
        dummy_config = cls.parse_obj({**template, **kwargs})
        return dummy_config

    @classmethod
    def list_profiles(cls) -> List[str]:
        "List the profiles contained in ``config_dir``."
        dummy_config = cls.get_dummy_config()
        profiles = [c.parent.name for c in dummy_config.config_dir.glob('profiles/*/config.json')]
        return profiles

    @classmethod
    def read(cls, profile: str) -> GlobalConfig:
        "Read the configuration from disk."
        dummy_config = cls.get_dummy_config(profile=profile)
        dummy_config.migrate_legacy_dirs()
        config = cls.parse_raw(dummy_config.config_file.read_text(encoding='utf-8'))
        if dummy_config.profile != config.profile:
            raise ValueError(
                'profile location does not match profile value of '
                f'"{config.profile}" in {dummy_config.config_file}'
            )
        return config

    def ensure_dirs(self) -> GlobalConfig:
        "Create the various folders used by instawow."
        for dir_ in (
            self.config_dir,
            self.profile_dir,
            self.logging_dir,
            self.plugin_dir,
            self.temp_dir,
            self.cache_dir,
        ):
            dir_.mkdir(exist_ok=True, parents=True)
        return self

    def write(self) -> GlobalConfig:
        """Write the configuration on disk.

        ``write``, unlike ``ensure_dirs``, should only be called when configuring
        instawow.  This means that environment overrides should only be persisted
        if made during configuration.
        """
        self.ensure_dirs()
        includes = {'addon_dir', 'game_flavour', 'profile'}
        output = self.json(include=includes, indent=2)
        self.config_file.write_text(output, encoding='utf-8')
        return self

    def migrate_legacy_dirs(self) -> GlobalConfig:
        "Migrate a profile-less configuration to the new format."
        legacy_config_file = self.config_dir / 'config.json'
        if (
            self.profile == self.__field_defaults__['profile']
            and not self.profile_dir.exists()
            and legacy_config_file.exists()
        ):
            legacy_json = json.loads(legacy_config_file.read_text(encoding='utf-8'))
            legacy_json.pop('profile', None)
            legacy_config = self.parse_obj(legacy_json)
            ignores = ignore_patterns('profiles')

            logger.info('migrating legacy configuration')
            copytree(self.config_dir, self.profile_dir, ignore=ignores)
            legacy_config.write()
            trash(
                [i for i in self.config_dir.iterdir() if i.name != 'profiles'], dst=self.temp_dir
            )

        return self

    def delete(self) -> None:
        "Delete the configuration files associated with this profile."
        trash((self.ensure_dirs().profile_dir,), dst=self.temp_dir, missing_ok=True)

    @property
    def is_classic(self) -> bool:
        return self.game_flavour == 'classic'

    @property
    def is_retail(self) -> bool:
        return self.game_flavour == 'retail'

    @property
    def profile_dir(self) -> Path:
        return self.config_dir / 'profiles' / self.profile

    @property
    def logging_dir(self) -> Path:
        return self.profile_dir / 'logs'

    @property
    def plugin_dir(self) -> Path:
        return self.profile_dir / 'plugins'

    @property
    def config_file(self) -> Path:
        return self.profile_dir / 'config.json'

    @property
    def db_file(self) -> Path:
        return self.profile_dir / 'db.sqlite'

    @property
    def cache_dir(self) -> Path:
        return self.temp_dir / 'cache'


Config = GlobalConfig


def setup_logging(config: GlobalConfig, log_level: Union[int, str] = 'INFO') -> int:
    import logging

    class InterceptHandler(logging.Handler):
        def emit(self, record: logging.LogRecord):
            # Get the corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where the logged message originated
            frame = logging.currentframe()
            depth = 2
            while frame and frame.f_code.co_filename == getattr(logging, '__file__', None):
                frame = frame.f_back
                depth += 1

            logger.opt(
                depth=depth,
                exception=record.exc_info,  # type: ignore
            ).log(level, record.getMessage())

    logging.basicConfig(handlers=[InterceptHandler()], level=log_level)
    handler = {
        'sink': config.logging_dir / 'error.log',
        'level': log_level,
        'rotation': '1 MB',
        'enqueue': True,
    }
    (handler_id,) = logger.configure(handlers=(handler,))
    return handler_id



from __future__ import annotations

from datetime import datetime, timezone
import enum
from itertools import chain, count, takewhile
import re
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    ClassVar,
    Dict,
    List,
    Optional as O,
    Sequence,
    Set,
    Union,
    cast,
)

from pydantic import BaseModel, WrongConstantError, validator
from typing_extensions import Literal, TypedDict, get_args
from yarl import URL

from . import exceptions as E, models as m
from .utils import bucketise, cached_property, gather, run_in_thread as t, slugify, uniq

if TYPE_CHECKING:
    from pydantic.fields import ModelField as ModeFieldT

    from .manager import Manager as ManagerT


class Strategies(enum.Enum):
    default = 'default'
    latest = 'latest'
    curse_latest_beta = 'curse_latest_beta'
    curse_latest_alpha = 'curse_latest_alpha'
    any_flavour = 'any_flavour'
    version = 'version'


class HashableModel(BaseModel):
    class Config:
        allow_mutation = False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        return hash(tuple(self.__dict__.items()))


class BaseStrategy(HashableModel):
    @validator('type_', check_fields=False, pre=True)
    def _parse_strategy(cls, value: Any, field: ModeFieldT) -> Strategies:
        accepted_strategies = get_args(field.type_)
        strategy = Strategies.__members__.get(value) or value
        if strategy in accepted_strategies:
            return strategy
        raise WrongConstantError(given=value, permitted=accepted_strategies)


class SimpleStrategy(BaseStrategy):
    type_: Literal[
        Strategies.default,
        Strategies.latest,
        Strategies.curse_latest_beta,
        Strategies.curse_latest_alpha,
        Strategies.any_flavour,
    ]


class VersionStrategy(BaseStrategy):
    type_: Literal[Strategies.version]
    version: str


class Defn(HashableModel):
    source: str
    id: O[str] = None
    alias: str
    strategy: Union[SimpleStrategy, VersionStrategy] = SimpleStrategy(type_=Strategies.default)

    @classmethod
    def from_pkg(cls, pkg: m.Pkg) -> Defn:
        strategy_kwargs: Dict[str, str] = (
            {'version': pkg.version} if pkg.options.strategy == 'version' else {}
        )
        return cls(
            source=pkg.source,
            id=pkg.id,
            alias=pkg.slug,
            strategy={'type_': pkg.options.strategy, **strategy_kwargs},
        )

    @classmethod
    def get(cls, source: str, alias: str) -> Defn:
        return cls(source=source, alias=alias)

    def with_(self, **kwargs: Any) -> Defn:
        return self.__class__(**{**self.__dict__, **kwargs})

    def with_strategy(self, strategy: Strategies) -> Defn:
        return self.with_(strategy={'type_': strategy})

    def with_version(self, version: str) -> Defn:
        return self.with_(strategy={'type_': Strategies.version, 'version': version})

    def __str__(self) -> str:
        return f'{self.source}:{self.alias}'


def normalise_names():
    import string

    trans_table = str.maketrans(dict.fromkeys(string.punctuation, ' '))

    def normalise(value: str):
        return ' '.join(value.casefold().translate(trans_table).split())

    return normalise


class _CatalogueEntryDefaultFields(TypedDict):
    source: str
    id: str
    slug: str
    name: str
    game_compatibility: Set[Literal['retail', 'classic']]
    folders: Sequence[Sequence[str]]
    download_count: int
    last_updated: Any


class _CatalogueEntry(BaseModel):
    source: str
    id: str
    slug: str
    name: str
    game_compatibility: Set[Literal['retail', 'classic']]
    folders: List[List[str]]
    download_count: int
    last_updated: datetime
    normalised_name: str
    derived_download_score: float


class MasterCatalogue(BaseModel):
    __root__: List[_CatalogueEntry]

    class Config:
        keep_untouched = (cast(Any, cached_property),)

    @classmethod
    async def collate(cls, age_cutoff: O[datetime]) -> MasterCatalogue:
        from types import SimpleNamespace

        from .manager import init_web_client

        resolvers = (CurseResolver, WowiResolver, TukuiResolver)

        async with init_web_client() as web_client:
            faux_manager = cast('ManagerT', SimpleNamespace(web_client=web_client))
            items = [a for r in resolvers async for a in r(faux_manager).collect_items()]

        normalise = normalise_names()
        most_downloads_per_source = {
            s: max(e['download_count'] for e in i)
            for s, i in bucketise(items, key=lambda v: v['source']).items()
        }
        entries = (
            _CatalogueEntry(
                **i,
                normalised_name=normalise(i['name']),
                derived_download_score=i['download_count']
                / most_downloads_per_source[i['source']],
            )
            for i in items
        )
        if age_cutoff:
            entries = (e for e in entries if e.last_updated >= age_cutoff)
        catalogue = cls.parse_obj(list(entries))
        return catalogue

    @cached_property
    def curse_slugs(self) -> Dict[str, str]:
        return {a.slug: a.id for a in self.__root__ if a.source == 'curse'}


class _OrmModel(BaseModel):
    class Config:
        allow_mutation = False
        orm_mode = True


class PkgModel_PkgFolder(_OrmModel):
    name: str


class PkgModel_PkgOptions(_OrmModel):
    strategy: str


class PkgModel_PkgDep(_OrmModel):
    id: str


class PkgModel_PkgVersion(_OrmModel):
    version: str
    install_time: datetime


class PkgModel(_OrmModel):
    source: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: datetime
    version: str
    folders: List[PkgModel_PkgFolder]
    options: PkgModel_PkgOptions
    deps: List[PkgModel_PkgDep]
    logged_versions: List[PkgModel_PkgVersion]


class MultiPkgModel(BaseModel):
    __root__: List[PkgModel]


class Resolver:
    source: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategies]]

    def __init__(self, manager: ManagerT) -> None:
        self.manager = manager

    def __init_subclass__(cls) -> None:
        async def resolve_wrapper(self: Resolver, defn: Defn, metadata: O[Any]) -> m.Pkg:
            if defn.strategy.type_ in self.strategies:
                return await resolve_one(self, defn, metadata)  # type: ignore
            raise E.PkgStrategyUnsupported(defn.strategy.type_)

        resolve_one = cls.resolve_one
        cls.resolve_one = resolve_wrapper  # type: ignore

    @property
    def supports_rollback(self) -> bool:
        "Whether the resolver supports rollback operations."
        return Strategies.version in self.strategies

    @staticmethod
    def get_alias_from_url(value: str) -> O[str]:
        "Attempt to extract a definition name from a given URL."

    async def resolve(self, defns: Sequence[Defn]) -> Dict[Defn, Any]:
        "Resolve add-on definitions into packages."
        results = await gather(self.resolve_one(d, None) for d in defns)
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[Any]) -> m.Pkg:
        "Resolve an individual definition into a package."
        raise NotImplementedError

    async def collect_items(self) -> AsyncIterable[_CatalogueEntryDefaultFields]:
        "Yield add-ons from source for cataloguing."
        return
        yield


if TYPE_CHECKING:

    # Only documenting the fields we're actually using -
    # the API responses are absolutely massive

    class CurseAddon_FileDependency(TypedDict):
        id: int  # Unique dependency ID
        addonId: int  # The ID of the add-on we're depending on
        # The type of dependency.  One of:
        #   1 = embedded library
        #   2 = optional dependency
        #   3 = required dependency (this is the one we're after)
        #   4 = tool
        #   5 = incompatible
        #   6 = include (wat)
        type: int
        fileId: int  # The ID of the parent file which has this as a dependency

    class CurseAddon_FileModules(TypedDict):
        foldername: str
        fingerprint: int  # Folder fingerprint used by Curse for reconciliation
        # One of:
        #   1 = package
        #   2 = module
        #   3 = main module
        #   4 = file
        #   5 = referenced file
        # For WoW add-ons the main folder will have type "3" and the rest
        # of them type "2"
        type: int

    class CurseAddon_File(TypedDict):
        id: int  # Unique file ID
        displayName: str  # Tends to be the version
        downloadUrl: str
        fileDate: str  # Upload datetime in ISO, e.g. '2020-02-02T12:12:12Z'
        releaseType: int  # 1 = stable; 2 = beta; 3 = alpha
        dependencies: List[CurseAddon_FileDependency]
        modules: List[CurseAddon_FileModules]
        exposeAsAlternative: O[bool]
        gameVersion: List[str]  # e.g. '8.3.0'
        gameVersionFlavor: Literal['wow_classic', 'wow_retail']

    class CurseAddon(TypedDict):
        id: int
        name: str  # User-facing add-on name
        websiteUrl: str  # e.g. 'https://www.curseforge.com/wow/addons/molinari'
        summary: str  # One-line description of the add-on
        downloadCount: int  # Total number of downloads
        latestFiles: List[CurseAddon_File]
        slug: str  # URL slug; 'molinari' in 'https://www.curseforge.com/wow/addons/molinari'
        dateReleased: str  # ISO datetime of latest release


class CurseResolver(Resolver):
    source = 'curse'
    name = 'CurseForge'
    strategies = {
        Strategies.default,
        Strategies.latest,
        Strategies.curse_latest_beta,
        Strategies.curse_latest_alpha,
        Strategies.any_flavour,
        Strategies.version,
    }

    # Reference: https://twitchappapi.docs.apiary.io/
    addon_api_url = URL('https://addons-ecs.forgesvc.net/api/v2/addon')

    @staticmethod
    def get_alias_from_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'www.wowace.com' and len(url.parts) > 2 and url.parts[1] == 'projects':
            return url.parts[2].lower()
        elif (
            url.host == 'www.curseforge.com'
            and len(url.parts) > 3
            and url.parts[1:3] == ('wow', 'addons')
        ):
            return url.parts[3].lower()

    async def resolve(self, defns: Sequence[Defn]) -> Dict[Defn, Any]:
        from aiohttp import ClientResponseError

        from .manager import cache_json_response

        defns_to_ids = {
            d: (d.id or self.manager.catalogue.curse_slugs.get(d.alias) or d.alias) for d in defns
        }
        numeric_ids = {i for i in defns_to_ids.values() if i.isdigit()}
        try:
            json_response = await cache_json_response(
                self.manager,
                self.addon_api_url,
                300,
                request_kwargs={'method': 'POST', 'json': list(numeric_ids)},
            )
        except ClientResponseError as error:
            if error.status != 404:
                raise
            json_response = []

        api_results = {str(r['id']): r for r in json_response}
        results = await gather(
            self.resolve_one(d, api_results.get(i)) for d, i in defns_to_ids.items()
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[CurseAddon]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent

        if defn.strategy.type_ is Strategies.version:

            from .manager import cache_json_response

            files = await cache_json_response(
                self.manager,
                self.addon_api_url / str(metadata['id']) / 'files',
                3600,
                label=f'Fetching metadata from {self.name}',
            )

            def is_compatible(f: CurseAddon_File):
                return cast(VersionStrategy, defn.strategy).version == f['displayName']

        else:

            files = metadata['latestFiles']

            def is_not_libless(f: CurseAddon_File):
                # There's also an 'isAlternate' field that's missing from some
                # 50 lib-less files from c. 2008.  'exposeAsAlternative' is
                # absent from the /file endpoint
                return not f['exposeAsAlternative']

            if defn.strategy.type_ is Strategies.any_flavour:

                def supports_game_version(f: CurseAddon_File):
                    return True

            else:
                classic_version_prefix = '1.13'
                flavour = 'wow_classic' if self.manager.config.is_classic else 'wow_retail'

                def supports_game_version(f: CurseAddon_File):
                    # Files can belong both to retail and classic
                    # but ``gameVersionFlavor`` can only be one of
                    # 'wow_retail' or 'wow_classic'.  To spice things up,
                    # ``gameVersion`` might not be populated so we still have to
                    # check the value of ``gameVersionFlavor``
                    return f['gameVersionFlavor'] == flavour or any(
                        v.startswith(classic_version_prefix) is self.manager.config.is_classic
                        for v in f['gameVersion']
                    )

            if defn.strategy.type_ is Strategies.latest:

                def has_release_type(f: CurseAddon_File):
                    return True

            elif defn.strategy.type_ is Strategies.curse_latest_beta:

                def has_release_type(f: CurseAddon_File):
                    return f['releaseType'] == 2

            elif defn.strategy.type_ is Strategies.curse_latest_alpha:

                def has_release_type(f: CurseAddon_File):
                    return f['releaseType'] == 3

            else:

                def has_release_type(f: CurseAddon_File):
                    return f['releaseType'] == 1

            def is_compatible(f: CurseAddon_File):
                return is_not_libless(f) and supports_game_version(f) and has_release_type(f)

        if not files:
            raise E.PkgFileUnavailable('no files available for download')
        try:
            file = max(
                filter(is_compatible, files),
                # The ``id`` is just a counter so we don't have to go digging around dates
                key=lambda f: f['id'],
            )
        except ValueError:
            raise E.PkgFileUnavailable(
                f'no files compatible with {self.manager.config.game_flavour} '
                f'using {defn.strategy.type_.name!r} strategy'
            )

        return m.Pkg(
            source=self.source,
            id=metadata['id'],
            slug=metadata['slug'],
            name=metadata['name'],
            description=metadata['summary'],
            url=metadata['websiteUrl'],
            download_url=file['downloadUrl'],
            date_published=file['fileDate'],
            version=file['displayName'],
            options=m.PkgOptions(strategy=defn.strategy.type_.name),
            deps=[m.PkgDep(id=d['addonId']) for d in file['dependencies'] if d['type'] == 3],
        )

    async def collect_items(self) -> AsyncIterable[_CatalogueEntryDefaultFields]:
        classic_version_prefix = '1.13'
        flavours = ('retail', 'classic')

        def excise_compatibility(files: List[CurseAddon_File]):
            for c in flavours:
                if any(f['gameVersionFlavor'] == f'wow_{c}' for f in files):
                    yield c
                elif any(
                    v.startswith(classic_version_prefix) is (c == 'classic')
                    for f in files
                    for v in f['gameVersion']
                ):
                    yield c

        step = 1000
        sort_order = '3'  # Alphabetical
        for index in count(0, step):
            url = (self.addon_api_url / 'search').with_query(
                gameId='1', sort=sort_order, pageSize=step, index=index
            )
            async with self.manager.web_client.get(url) as response:
                json_response: List[CurseAddon] = await response.json()

            if not json_response:
                break
            for item in json_response:
                folders = uniq(
                    tuple(m['foldername'] for m in f['modules'])
                    for f in item['latestFiles']
                    if not f['exposeAsAlternative']
                )
                yield _CatalogueEntryDefaultFields(
                    source=self.source,
                    id=str(item['id']),
                    slug=item['slug'],
                    name=item['name'],
                    game_compatibility=set(excise_compatibility(item['latestFiles'])),
                    folders=folders,
                    download_count=item['downloadCount'],
                    last_updated=item['dateReleased'],
                )


class WowiCommonTerms(TypedDict):
    UID: str  # Unique add-on ID
    UICATID: str  # ID of category add-on is placed in
    UIVersion: str  # Add-on version
    UIDate: int  # Upload date expressed as unix epoch
    UIName: str  # User-facing add-on name
    UIAuthorName: str


class WowiCompatibilityEntry(TypedDict):
    version: str  # Game version, e.g. '8.3.0'
    name: str  # Xpac or patch name, e.g. "Visions of N'Zoth" for 8.3.0


class WowiListApiItem(WowiCommonTerms):
    UIFileInfoURL: str  # Add-on page on WoWI
    UIDownloadTotal: str  # Total number of downloads
    UIDownloadMonthly: str  # Number of downloads in the last month and not 'monthly'
    UIFavoriteTotal: str
    UICompatibility: O[List[WowiCompatibilityEntry]]  # ``null`` if would be empty
    UIDir: List[str]  # Names of folders contained in archive
    UIIMG_Thumbs: O[List[str]]  # Thumbnail URLs; ``null`` if would be empty
    UIIMGs: O[List[str]]  # Full-size image URLs; ``null`` if would be empty
    # There are only two add-ons on the entire list with siblings
    # (they refer to each other). I don't know if this was meant to capture
    # dependencies (probably not) but it's so underused as to be worthless.
    # ``null`` if would be empty
    UISiblings: O[List[str]]
    UIDonationLink: O[str]  # Absent from the first item on the list (!)


class WowiDetailsApiItem(WowiCommonTerms):
    UIMD5: O[str]  # Archive hash, ``null` when UI is pending
    UIFileName: str  # The actual filename, e.g. 'foo.zip'
    UIDownload: str  # Download URL
    UIPending: Literal['0', '1']  # Set to '1' if the file is awaiting approval
    UIDescription: str  # Long description with BB Code and all
    UIChangeLog: str  # This can also contain BB Code
    UIHitCount: str  # Same as UIDownloadTotal
    UIHitCountMonthly: str  # Same as UIDownloadMonthly


class WowiCombinedItem(WowiListApiItem, WowiDetailsApiItem):
    pass


class WowiResolver(Resolver):
    source = 'wowi'
    name = 'WoWInterface'
    strategies = {Strategies.default}

    # Reference: https://api.mmoui.com/v3/globalconfig.json
    # There's also a v4 API corresponding to the as yet unreleased Minion v4,
    # which is fair to assume is unstable.  They changed the naming scheme to
    # camelCase and some fields which were strings were converted to numbers.
    # Neither API provides access to classic files for multi-file add-ons and
    # 'UICompatibility' can't be relied on to enforce compatibility
    # in instawow.  The API appears to inherit the version of the latest
    # file to have been uploaded, which for multi-file add-ons can be the
    # classic version.  Hoooowever the download link always points to the
    # 'retail' version, which for single-file add-ons belonging to the
    # classic category would be an add-on for classic.
    list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    list_api_items: O[Dict[str, WowiListApiItem]] = None

    @staticmethod
    def get_alias_from_url(value: str) -> O[str]:
        url = URL(value)
        if (
            url.host in {'wowinterface.com', 'www.wowinterface.com'}
            and len(url.parts) == 3
            and url.parts[1] == 'downloads'
        ):
            if url.name == 'landing.php':
                source_id = url.query.get('fileid')
                if source_id:
                    return source_id
            elif url.name == 'fileinfo.php':
                source_id = url.query.get('id')
                if source_id:
                    return source_id
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)', url.name)
                if match:
                    return match.group('id')

    async def resolve(self, defns: Sequence[Defn]) -> Dict[Defn, Any]:
        from aiohttp import ClientResponseError

        from .manager import cache_json_response

        async with self.manager.locks['load WoWI catalogue']:
            if self.list_api_items is None:
                list_api_items = await cache_json_response(
                    self.manager,
                    self.list_api_url,
                    3600,
                    label=f'Synchronising {self.name} catalogue',
                )
                self.list_api_items = {i['UID']: i for i in list_api_items}

        defns_to_ids = {d: ''.join(takewhile(str.isdigit, d.alias)) for d in defns}
        numeric_ids = {i for i in defns_to_ids.values() if i.isdigit()}
        url = self.details_api_url / f'{",".join(numeric_ids)}.json'
        try:
            details_api_items = await cache_json_response(self.manager, url, 300)
        except ClientResponseError as error:
            if error.status != 404:
                raise
            details_api_items = []

        combined_items = {
            r['UID']: WowiCombinedItem(**{**self.list_api_items[r['UID']], **r})
            for r in details_api_items
        }
        results = await gather(
            self.resolve_one(d, combined_items.get(i)) for d, i in defns_to_ids.items()
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[WowiCombinedItem]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent

        return m.Pkg(
            source=self.source,
            id=metadata['UID'],
            slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
            name=metadata['UIName'],
            description=metadata['UIDescription'],
            url=metadata['UIFileInfoURL'],
            download_url=metadata['UIDownload'],
            date_published=metadata['UIDate'],
            version=metadata['UIVersion'],
            options=m.PkgOptions(strategy=defn.strategy.type_.name),
        )

    async def collect_items(self) -> AsyncIterable[_CatalogueEntryDefaultFields]:
        async with self.manager.web_client.get(self.list_api_url) as response:
            list_api_items: List[WowiListApiItem] = await response.json()

        for list_item in list_api_items:
            yield _CatalogueEntryDefaultFields(
                source=self.source,
                id=list_item['UID'],
                name=list_item['UIName'],
                slug='',
                folders=[list_item['UIDir']],
                game_compatibility={'classic', 'retail'},
                download_count=int(list_item['UIDownloadTotal']),
                last_updated=list_item['UIDate'],
            )


if TYPE_CHECKING:

    class TukuiUi(TypedDict):
        author: str
        category: str
        changelog: str
        donate_url: str
        downloads: int
        git: str
        id: Literal[-1, -2]  # -1 is Tukui and -2 ElvUI
        lastdownload: str
        lastupdate: str  # ISO date and no tz, e.g. '2020-02-02'
        name: str
        patch: str
        screenshot_url: str
        small_desc: str
        ticket: str
        url: str
        version: str
        web_url: str

    class TukuiAddon(TypedDict):
        author: str
        category: str
        changelog: str
        donate_url: str
        downloads: str  # Not a mistake, it is actually a string
        id: str
        last_download: str
        # ISO *datetime* with space sep and without an offset, e.g. '2020-02-02 12:12:20'
        lastupdate: str
        name: str
        patch: str
        screenshot_url: str
        small_desc: str
        url: str
        version: str
        web_url: str


class TukuiResolver(Resolver):
    source = 'tukui'
    name = 'Tukui'
    strategies = {Strategies.default}

    # There's also a ``/client-api.php`` endpoint which is apparently
    # used by the Tukui client itself to check for updates for the two retail
    # UIs only.  The response body appears to be identical to ``/api.php``
    api_url = URL('https://www.tukui.org/api.php')

    retail_uis = {'-1': 'tukui', '-2': 'elvui', 'tukui': 'tukui', 'elvui': 'elvui'}

    @staticmethod
    def get_alias_from_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'www.tukui.org' and url.path in {
            '/addons.php',
            '/classic-addons.php',
            '/download.php',
        }:
            alias = url.query.get('id') or url.query.get('ui')
            if alias:
                return alias

    async def resolve_one(self, defn: Defn, metadata: Any) -> m.Pkg:
        alias = ui_id = self.retail_uis.get(defn.alias)
        if not alias:
            alias = ''.join(takewhile('-'.__ne__, defn.alias))

        if self.manager.config.is_classic:
            query = 'classic-addon'
        elif ui_id:
            query = 'ui'
        else:
            query = 'addon'

        url = self.api_url.with_query({query: alias})
        async with self.manager.web_client.get(url) as response:
            # Does not 404 for non-existent add-ons but the response body is empty
            if not response.content_length:
                raise E.PkgNonexistent
            addon: Union[TukuiUi, TukuiAddon] = await response.json(content_type=None)  # text/html

        return m.Pkg(
            source=self.source,
            id=addon['id'],
            slug=ui_id or slugify(f'{addon["id"]} {addon["name"]}'),
            name=addon['name'],
            description=addon['small_desc'],
            url=addon['web_url'],
            download_url=addon['url'],
            date_published=datetime.fromisoformat(addon['lastupdate']).replace(
                tzinfo=timezone.utc
            ),
            version=addon['version'],
            options=m.PkgOptions(strategy=defn.strategy.type_.name),
        )

    async def collect_items(self) -> AsyncIterable[_CatalogueEntryDefaultFields]:
        for query, param in [
            ('ui', 'tukui'),
            ('ui', 'elvui'),
            ('addons', 'all'),
            ('classic-addons', 'all'),
        ]:
            async with self.manager.web_client.get(
                self.api_url.with_query({query: param})
            ) as response:
                metadata = await response.json(content_type=None)  # text/html

            items: List[Union[TukuiUi, TukuiAddon]] = [metadata] if query == 'ui' else metadata
            game_compatibility = 'classic' if query == 'classic-addons' else 'retail'
            for item in items:
                yield _CatalogueEntryDefaultFields(
                    source=self.source,
                    id=str(item['id']),
                    slug='',
                    name=item['name'],
                    folders=[],
                    game_compatibility={game_compatibility},
                    # Split Tukui and ElvUI downloads evenly between them.
                    # They both have the exact same number of downloads so
                    # I'm assuming they're being counted together.
                    # Anyway, this should help with scoring other add-ons
                    # on the Tukui catalogue higher
                    download_count=int(item['downloads']) // (2 if query == 'ui' else 1),
                    last_updated=datetime.fromisoformat(item['lastupdate']).replace(
                        tzinfo=timezone.utc
                    ),
                )


if TYPE_CHECKING:

    # Not exhaustive (as you might've guessed).  Reference:
    # https://docs.github.com/en/rest/reference/repos

    class GithubRepo(TypedDict):
        name: str  # the repo in user-or-org/repo
        full_name: str  # user-or-org/repo
        description: str
        html_url: str

    class GithubReleaseAsset(TypedDict):
        name: str  # filename
        content_type: str  # mime type
        state: Literal['starter', 'uploaded']
        browser_download_url: str

    class GithubRelease(TypedDict):
        tag_name: str  # Hopefully the version
        published_at: str  # ISO datetime
        assets: List[GithubReleaseAsset]


class GithubResolver(Resolver):
    source = 'github'
    name = 'GitHub'
    strategies = {
        Strategies.default,
        Strategies.version,
    }

    repos_api_url = URL('https://api.github.com/repos')

    @staticmethod
    def get_alias_from_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'github.com' and len(url.parts) > 2:
            return '/'.join(url.parts[1:3])

    async def resolve_one(self, defn: Defn, metadata: Any) -> m.Pkg:
        """Resolve a hypothetical add-on hosted on GitHub.

        The GitHub resolver is inspired by strongbox's
        (see https://github.com/ogri-la/strongbox/blob/develop/github-addons.md)
        and makes similar assumptions.
        The repo must have releases.  The releases must have assets
        attached to them.  instawow will not retrieve add-ons from VCS tarballs
        or 'zipballs' (i.e. from source).
        It will prioritise assets which appear to be compatible with the
        selected game flavour.
        It will *not* look for TOC files or validate the contents
        of the ZIP file in any way.
        """
        from aiohttp import ClientResponseError

        from .manager import cache_json_response

        repo_url = self.repos_api_url / defn.alias
        try:
            project_metadata: GithubRepo = await cache_json_response(self.manager, repo_url, 3600)
        except ClientResponseError as error:
            if error.status == 404:
                raise E.PkgNonexistent
            raise

        if defn.strategy.type_ is Strategies.version:
            release_url = repo_url / 'releases/tags' / cast(VersionStrategy, defn.strategy).version
        else:
            release_url = repo_url / 'releases/latest'
        async with self.manager.web_client.get(release_url) as response:
            if response.status == 404:
                raise E.PkgFileUnavailable('release not found')
            release_metadata: GithubRelease = await response.json()

        try:

            def is_valid_asset(asset: GithubReleaseAsset):
                return (
                    # There is something of a convention that Classic archives
                    # end in '-classic' and lib-less archives end in '-nolib'.
                    # Archives produced by packager follow this convention
                    not asset['name'].endswith(('-nolib.zip', '-nolib-classic.zip'))
                    and asset['content_type']
                    in {'application/zip', 'application/x-zip-compressed'}
                    # A failed upload has a state value of 'starter'
                    and asset['state'] == 'uploaded'
                )

            assets = release_metadata['assets']
            matching_asset = next(
                chain(
                    (
                        a
                        for a in assets
                        if is_valid_asset(a)
                        and a['name'].endswith('-classic.zip') is self.manager.config.is_classic
                    ),
                    filter(is_valid_asset, assets),
                )
            )
        except StopIteration:
            raise E.PkgFileUnavailable

        return m.Pkg(
            source=self.source,
            id=project_metadata['full_name'],
            slug=project_metadata['full_name'].lower(),
            name=project_metadata['name'],
            description=project_metadata['description'],
            url=project_metadata['html_url'],
            download_url=matching_asset['browser_download_url'],
            date_published=release_metadata['published_at'],
            version=release_metadata['tag_name'],
            options=m.PkgOptions(strategy=defn.strategy.type_.name),
        )


class InstawowResolver(Resolver):
    source = 'instawow'
    name = 'instawow'
    strategies = {Strategies.default}

    _addons = {
        ('0', 'weakauras-companion'),
        ('1', 'weakauras-companion-autoupdate'),
    }

    async def resolve_one(self, defn: Defn, metadata: Any) -> m.Pkg:
        try:
            source_id, slug = next(p for p in self._addons if defn.alias in p)
        except StopIteration:
            raise E.PkgNonexistent

        from .wa_updater import BuilderConfig, WaCompanionBuilder

        sentinel = '__sentinel__'
        builder = WaCompanionBuilder(self.manager, BuilderConfig(account=sentinel))
        if source_id == '1':
            if builder.builder_config.account == sentinel:
                raise E.PkgFileUnavailable('account name not provided')
            await builder.build()

        checksum = await t(builder.checksum)()
        return m.Pkg(
            source=self.source,
            id=source_id,
            slug=slug,
            name='WeakAuras Companion',
            description='A WeakAuras Companion clone.',
            url='https://github.com/layday/instawow',
            download_url=builder.addon_file.as_uri(),
            date_published=datetime.now(),
            version=checksum[:7],
            options=m.PkgOptions(strategy=defn.strategy.type_.name),
        )



from __future__ import annotations

from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional as O,
    Sequence,
    Tuple,
    Type,
    cast,
)

from loguru import logger
from pydantic import BaseModel, Field, validator
from typing_extensions import Literal, TypedDict
from yarl import URL

from .config import BaseConfig, Config as GlobalConfig
from .utils import bucketise, chain_dict, gather, iter_in_thread as iit, run_in_thread as t

if TYPE_CHECKING:
    from pathlib import Path

    from .manager import Manager

    ImportString = str
    RemoteAura = Tuple[List[WeakAura], WagoApiResponse, ImportString]


import_api = URL('https://data.wago.io/api/raw/encoded')


class BuilderConfig(BaseConfig):
    account: str
    api_key: O[str]

    class Config:  # type: ignore
        env_prefix = 'WAC_'

    def get_saved_vars(self, global_config: GlobalConfig) -> Path:
        root_for_flavour = global_config.addon_dir.parents[1]
        return root_for_flavour / 'WTF' / 'Account' / self.account / 'SavedVariables'


class WeakAura(BaseModel):
    id: str
    uid: str
    parent: O[str]
    url: URL
    version: int

    class Config:
        arbitrary_types_allowed = True

    _convert_url = validator('url', pre=True)(lambda v: URL(v))


class WeakAuras:
    filename = 'WeakAuras.lua'
    table_prefix = 'WeakAurasSaved'
    api = URL('https://data.wago.io/api/check/weakauras')

    entries: Dict[str, List[WeakAura]]

    def __init__(self, lua_table: Dict[Any, Any]) -> None:
        auras = (WeakAura.parse_obj(a) for a in lua_table['displays'].values() if a.get('url'))
        self.entries = bucketise(auras, key=lambda a: a.url.parts[1])


class Plateroo(WeakAura):
    id: str = Field(alias='Name')
    uid = ''


class Plateroos(WeakAuras):
    filename = 'Plater.lua'
    table_prefix = 'PlaterDB'
    api = URL('https://data.wago.io/api/check/plater')

    def __init__(self, lua_table: Dict[Any, Any]) -> None:
        auras = (
            Plateroo.parse_obj(a)
            for n, p in lua_table['profiles'].items()
            for a in chain(
                ({**p, 'Name': f'__profile_{n}__'},),
                (i for n, v in p.items() if n in {'script_data', 'hook_data'} for i in v),
            )
            if a.get('url')
        )
        self.entries = {a.url.parts[1]: [cast(WeakAura, a)] for a in auras}


if TYPE_CHECKING:

    class WagoApiChangelog(TypedDict, total=False):
        format: Literal['bbcode', 'markdown']
        text: str

    class WagoApiCommonFields(TypedDict):
        _id: str  # +   # Alphanumeric ID
        name: str  # +  # User-facing name
        slug: str  # +  # Slug if it has one; otherwise same as ``_id``
        url: str
        created: str  # ISO datetime
        modified: str  # ISO datetime
        game: str  # "classic" or xpac, e.g. "bfa"
        username: str  # +  # Author username
        version: int  # +   # Version counter, incremented with every update
        # Semver auto-generated from ``version`` - for presentation only
        versionString: str  # +
        changelog: WagoApiChangelog  # +

    class WagoApiOptionalFields(TypedDict, total=False):
        forkOf: str  # Only present on forks
        regionType: str  # Only present on WAs

    class WagoApiResponse(WagoApiCommonFields, WagoApiOptionalFields):
        pass


class WaCompanionBuilder:
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: Manager, builder_config: BuilderConfig) -> None:
        self.manager = manager
        self.addon_file = self.manager.config.plugin_dir / __name__ / 'WeakAurasCompanion.zip'
        self.builder_config = builder_config

    @staticmethod
    def extract_auras(model: Type[WeakAuras], source: str) -> WeakAuras:
        import re

        from slpp import SLPP

        class WaParser(SLPP):
            def decode(self, text: str) -> Any:
                text = re.sub(rf'^\s*{model.table_prefix} = ', '', text)
                text = re.sub(r' -- \[\d+\]$', '', text, flags=re.M)
                self.text = text
                self.at, self.ch, self.depth = 0, '', 0
                self.len = len(text)
                self.next_chr()
                return self.value()  # type: ignore

        table = WaParser().decode(source)
        return model(table)

    def extract_installed_auras(self) -> Iterator[WeakAuras]:
        import time

        saved_vars = self.builder_config.get_saved_vars(self.manager.config)
        for model in WeakAuras, Plateroos:
            file = saved_vars / model.filename
            if not file.exists():
                logger.info(f'{file} not found')
            else:
                start = time.perf_counter()
                aura_groups = self.extract_auras(
                    model, file.read_text(encoding='utf-8-sig', errors='replace')
                )
                logger.debug(f'{model.__name__} extracted in {time.perf_counter() - start}s')
                yield aura_groups

    async def get_wago_metadata(self, aura_groups: WeakAuras) -> List[WagoApiResponse]:
        aura_ids = list(aura_groups.entries)
        url = aura_groups.api.with_query(ids=','.join(aura_ids))
        async with self.manager.web_client.get(
            url, headers={'api-key': self.builder_config.api_key or ''}
        ) as response:
            metadata: List[WagoApiResponse] = await response.json()

        results = chain_dict(aura_ids, None, ((i['slug'], i) for i in metadata))
        return list(results.values())

    async def get_wago_import_string(self, aura_id: str) -> str:
        async with self.manager.web_client.get(
            import_api.with_query(id=aura_id),
            headers={'api-key': self.builder_config.api_key or ''},
        ) as response:
            return await response.text()

    async def get_remote_auras(
        self, aura_groups: WeakAuras
    ) -> Tuple[Type[WeakAuras], List[RemoteAura]]:
        if not aura_groups.entries:
            return (aura_groups.__class__, [])

        metadata = await self.get_wago_metadata(aura_groups)
        import_strings = await gather(
            (self.get_wago_import_string(r['_id']) for r in metadata), False
        )
        return (
            aura_groups.__class__,
            list(filter(all, zip(aura_groups.entries.values(), metadata, import_strings))),
        )

    def make_addon(self, auras: Sequence[Tuple[type, List[RemoteAura]]]) -> None:
        from functools import partial
        from importlib.resources import read_text
        from zipfile import ZipFile, ZipInfo

        from jinja2 import Environment, FunctionLoader

        from . import wa_templates

        jinja_env = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FunctionLoader(partial(read_text, wa_templates)),
        )
        aura_dict: Dict[type, List[RemoteAura]] = chain_dict((WeakAuras, Plateroos), [], auras)

        self.addon_file.parent.mkdir(exist_ok=True)
        with ZipFile(self.addon_file, 'w') as file:

            def write_tpl(filename: str, ctx: Dict[str, Any]) -> None:
                # We're not using a plain string as the first argument to
                # ``writestr`` 'cause then the timestamp is generated dynamically
                # making the build unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                file.writestr(zip_info, jinja_env.get_template(filename).render(ctx))

            write_tpl(
                'data.lua',
                {
                    'weakauras': [
                        (
                            metadata['slug'],
                            {
                                'name': metadata['name'],
                                'author': metadata['username'],
                                'encoded': import_string,
                                'wagoVersion': metadata['version'],
                                'wagoSemver': metadata['versionString'],
                                'versionNote': metadata['changelog'].get('text', ''),
                            },
                        )
                        for _, metadata, import_string in aura_dict[WeakAuras]
                    ],
                    # Maps internal UIDs of top-level auras to IDs or slugs on Wago
                    'weakaura_uids': [
                        (a.uid, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in (
                            next((i for i in existing_auras if not i.parent), existing_auras[0]),
                        )
                    ],
                    # Maps local names to IDs or slugs on Wago
                    'weakaura_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in existing_auras
                    ],
                    'plateroos': [
                        (
                            metadata['slug'],
                            {
                                'name': metadata['name'],
                                'author': metadata['username'],
                                'encoded': import_string,
                                'wagoVersion': metadata['version'],
                                'wagoSemver': metadata['versionString'],
                                'versionNote': metadata['changelog'].get('text', ''),
                            },
                        )
                        for _, metadata, import_string in aura_dict[Plateroos]
                    ],
                    'plater_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[Plateroos]
                        for a in existing_auras
                    ],
                },
            )
            write_tpl('init.lua', {})
            write_tpl(
                'WeakAurasCompanion.toc',
                {'interface': '11305' if self.manager.config.is_classic else '80300'},
            )

    async def build(self) -> None:
        installed_auras = iit(self.extract_installed_auras())
        remote_auras = await gather(
            [self.get_remote_auras(g) async for g in installed_auras],
            False,
        )
        await t(self.make_addon)(remote_auras)

    def checksum(self) -> str:
        from hashlib import sha256

        return sha256(self.addon_file.read_bytes()).hexdigest()



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, List, Optional as O, Type

import pydantic
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKeyConstraint,
    String,
    TypeDecorator,
    and_,
    exc,
    func,
    inspect,
)
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
from sqlalchemy.orm import Session, object_session, relationship


class _BaseCoercer(pydantic.BaseModel):
    """The coercer is used inside the declarative constructor to type-cast
    values _prior to_ inserts.  SQLAlchemy delegates this function to
    the DB API -- see https://stackoverflow.com/a/8980982.  We need this to
    happen in advance to be able to compare values with their in-database
    counterparts.  It also saves us the trouble of having to manually
    parse dates and the like.
    """

    class Config:
        extra = pydantic.Extra.allow
        max_anystr_length = 2 ** 32


_coercers: Dict[type, Type[pydantic.BaseModel]] = {}


class _ModelMeta(DeclarativeMeta):
    def __init__(cls, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        inspector = inspect(cls, raiseerr=False)
        if inspector:
            _coercers[cls] = pydantic.create_model(
                f'{cls.__name__}Coercer',
                __base__=_BaseCoercer,
                **{
                    c.name: (c.type.python_type, ...)
                    for c in inspector.columns
                    if not c.foreign_keys and not c.name.startswith('_') and not c.server_default
                },
            )


def _constructor(self: object, **kwargs: Any) -> None:
    intermediate_obj = _coercers[self.__class__](**kwargs)
    for k, v in intermediate_obj:
        setattr(self, k, v)


ModelBase: Any = declarative_base(constructor=_constructor, metaclass=_ModelMeta)

if TYPE_CHECKING:
    TZDateTime_base_class = TypeDecorator[datetime]
else:
    TZDateTime_base_class = TypeDecorator


class TZDateTime(TZDateTime_base_class):
    impl = DateTime

    def process_bind_param(self, value: O[datetime], dialect: Any) -> O[datetime]:  # type: ignore
        if value is not None:
            if not value.tzinfo:
                raise TypeError('tzinfo is required')
            value = value.astimezone(timezone.utc).replace(tzinfo=None)
        return value

    def process_result_value(self, value: O[datetime], dialect: Any) -> O[datetime]:
        return value and value.replace(tzinfo=timezone.utc)

    @property
    def python_type(self):
        return datetime


class Pkg(ModelBase):
    __tablename__ = 'pkg'

    source = Column(String, primary_key=True)
    id = Column(String, primary_key=True)
    slug = Column(String, nullable=False)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    url = Column(String, nullable=False)
    download_url = Column(String, nullable=False)
    date_published = Column(TZDateTime, nullable=False)
    version = Column(String, nullable=False)
    folders: relationship[List[PkgFolder]] = relationship(
        'PkgFolder', cascade='all, delete-orphan', backref='pkg'
    )
    options: relationship[PkgOptions] = relationship(
        'PkgOptions', cascade='all, delete-orphan', uselist=False
    )
    deps: relationship[List[PkgDep]] = relationship(
        'PkgDep', cascade='all, delete-orphan', backref='pkg'
    )

    @property
    def logged_versions(self) -> List[PkgVersionLog]:
        session: O[Session] = object_session(self)
        return (
            (
                session.query(PkgVersionLog)
                .filter(
                    and_(PkgVersionLog.pkg_source == self.source, PkgVersionLog.pkg_id == self.id)
                )
                .order_by(PkgVersionLog.install_time.desc())
                .all()
            )
            if session
            else []
        )


class PkgFolder(ModelBase):
    __tablename__ = 'pkg_folder'
    __table_args__ = (ForeignKeyConstraint(['pkg_source', 'pkg_id'], ['pkg.source', 'pkg.id']),)

    name = Column(String, primary_key=True)
    pkg_source = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


class PkgOptions(ModelBase):
    __tablename__ = 'pkg_options'
    __table_args__ = (ForeignKeyConstraint(['pkg_source', 'pkg_id'], ['pkg.source', 'pkg.id']),)

    strategy = Column(String, nullable=False)
    pkg_source = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


class PkgDep(ModelBase):
    __tablename__ = 'pkg_dep'
    __table_args__ = (ForeignKeyConstraint(['pkg_source', 'pkg_id'], ['pkg.source', 'pkg.id']),)

    id = Column(String, primary_key=True)
    pkg_source = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


class PkgVersionLog(ModelBase):
    __tablename__ = 'pkg_version_log'

    version = Column(String, primary_key=True)
    install_time = Column(TZDateTime, nullable=False, server_default=func.now())
    pkg_source = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


def should_migrate(engine: Any, version: str) -> bool:
    """Check if the database version is the same as `version`;
    if not a migration would be required.
    """
    with engine.begin() as conn:
        try:
            current = conn.execute(
                'SELECT version_num FROM alembic_version WHERE version_num = (?)',
                version,
            ).scalar()
        except exc.OperationalError:
            return True
        else:
            return not current


def is_pkg(value: Any) -> bool:
    return isinstance(value, Pkg)



import enum
from functools import partial
import json

from click.testing import CliRunner
import pytest

from instawow import cli, config


class Flavour(enum.IntEnum):
    retail = enum.auto()
    classic = enum.auto()


@pytest.fixture(autouse=True)
def mock(mock_all):
    pass


@pytest.fixture(params=list(Flavour.__members__))
def cli_config(
    request,
    tmp_path_factory,
    temp_dir,
):
    "Reuse the temporary folder for all cases of a parametrized test."
    parametrized_tmp_path = (
        tmp_path_factory.getbasetemp()
        / f'{request.node.name[: request.node.name.index("[")]}_{request.param}'
    )
    parametrized_tmp_path.mkdir(exist_ok=True)
    addon_dir = parametrized_tmp_path / 'addons'
    addon_dir.mkdir(exist_ok=True)
    return config.Config(
        config_dir=parametrized_tmp_path / 'config',
        addon_dir=addon_dir,
        game_flavour=request.param,
        temp_dir=temp_dir,
    ).write()


@pytest.fixture
def run(
    monkeypatch,
    request,
    event_loop,
    cli_config,
):
    param = getattr(request, 'param', None)
    if param is Flavour.retail and cli_config.is_classic:
        pytest.skip('test is for retail only')
    elif param is Flavour.classic and cli_config.is_retail:
        pytest.skip('test is for classic only')

    monkeypatch.setattr('instawow.manager.CliManager.run', event_loop.run_until_complete)
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(cli_config.config_dir))
    yield partial(
        CliRunner().invoke,
        cli.main,
        catch_exceptions=False,
    )


@pytest.fixture
def molinari_and_run(run):
    run('install curse:molinari')
    yield run


@pytest.mark.parametrize(
    'args, cmp',
    [
        (
            'install curse:molinari',
            lambda v: v.startswith('✓ curse:molinari\n  installed'),
        ),
        (
            'install curse:molinari',
            '✗ curse:molinari\n  package already installed\n'.__eq__,
        ),
        (
            'update curse:molinari',
            '✗ curse:molinari\n  package is up to date\n'.__eq__,
        ),
        (
            'remove curse:molinari',
            '✓ curse:molinari\n  removed\n'.__eq__,
        ),
        (
            'update curse:molinari',
            '✗ curse:molinari\n  package is not installed\n'.__eq__,
        ),
        (
            'remove curse:molinari',
            '✗ curse:molinari\n  package is not installed\n'.__eq__,
        ),
    ],
)
def test_valid_curse_pkg_lifecycle(run, args, cmp):
    assert cmp(run(args).output)


@pytest.mark.parametrize(
    'args, cmp, run',
    [
        (
            'install tukui:1',
            lambda v: v.startswith('✓ tukui:1\n  installed'),
            None,
        ),
        (
            'install tukui:1',
            '✗ tukui:1\n  package already installed\n'.__eq__,
            None,
        ),
        (
            'update tukui:1',
            '✗ tukui:1-merathilisui\n  package is up to date\n'.__eq__,
            Flavour.retail,
        ),
        (
            'update tukui:1',
            '✗ tukui:1-tukui\n  package is up to date\n'.__eq__,
            Flavour.classic,
        ),
        (
            'remove tukui:1',
            '✓ tukui:1\n  removed\n'.__eq__,
            None,
        ),
        (
            'update tukui:1',
            '✗ tukui:1\n  package is not installed\n'.__eq__,
            None,
        ),
        (
            'remove tukui:1',
            '✗ tukui:1\n  package is not installed\n'.__eq__,
            None,
        ),
        (
            'install tukui:tukui',
            lambda v: v.startswith('✓ tukui:tukui\n  installed'),
            Flavour.retail,
        ),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package already installed\n'.__eq__,
            Flavour.retail,
        ),
        (
            'update tukui:tukui',
            '✗ tukui:tukui\n  package is up to date\n'.__eq__,
            Flavour.retail,
        ),
        (
            'remove tukui:tukui',
            '✓ tukui:tukui\n  removed\n'.__eq__,
            Flavour.retail,
        ),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package does not exist\n'.__eq__,
            Flavour.classic,
        ),
        (
            'update tukui:tukui',
            '✗ tukui:tukui\n  package is not installed\n'.__eq__,
            None,
        ),
        (
            'remove tukui:tukui',
            '✗ tukui:tukui\n  package is not installed\n'.__eq__,
            None,
        ),
    ],
    indirect=('run',),
)
def test_valid_tukui_pkg_lifecycle(run, args, cmp):
    assert cmp(run(args).output)


@pytest.mark.parametrize(
    'args, cmp',
    [
        (
            'install wowi:13188-molinari',
            lambda v: v.startswith('✓ wowi:13188-molinari\n  installed'),
        ),
        (
            'install wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package already installed\n'.__eq__,
        ),
        (
            'update wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package is up to date\n'.__eq__,
        ),
        (
            'remove wowi:13188-molinari',
            '✓ wowi:13188-molinari\n  removed\n'.__eq__,
        ),
        (
            'update wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package is not installed\n'.__eq__,
        ),
        (
            'remove wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package is not installed\n'.__eq__,
        ),
    ],
)
def test_valid_wowi_pkg_lifecycle(run, args, cmp):
    assert cmp(run(args).output)


@pytest.mark.parametrize(
    'args, cmp, run',
    [
        (
            'install curse:molinari',
            lambda v: v.startswith('✓ curse:molinari\n  installed'),
            Flavour.retail,
        ),
        (
            'install wowi:13188-molinari',
            "✗ wowi:13188-molinari\n"
            "  package folders conflict with installed package curse:molinari\n".__eq__,
            Flavour.retail,
        ),
        (
            'remove curse:molinari',
            '✓ curse:molinari\n  removed\n'.__eq__,
            Flavour.retail,
        ),
    ],
    indirect=('run',),
)
def test_folder_conflict_lifecycle(run, args, cmp):
    assert cmp(run(args).output)


def test_preexisting_folder_conflict_on_install(cli_config, run):
    cli_config.addon_dir.joinpath('Molinari').mkdir()
    assert (
        run('install curse:molinari').output
        == '''\
✗ curse:molinari
  package folders conflict with 'Molinari'
'''
    )
    assert run('install --replace curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed'
    )


@pytest.mark.parametrize(
    'args, output',
    [
        (
            'install curse:gargantuan-wigs',
            '✗ curse:gargantuan-wigs\n  package does not exist\n',
        ),
        (
            'update curse:gargantuan-wigs',
            '✗ curse:gargantuan-wigs\n  package is not installed\n',
        ),
        (
            'remove curse:gargantuan-wigs',
            '✗ curse:gargantuan-wigs\n  package is not installed\n',
        ),
    ],
)
def test_invalid_addon_name_lifecycle(run, args, output):
    assert run(args).output == output


@pytest.mark.parametrize(
    'args, output',
    [
        (
            'install foo:bar',
            '✗ foo:bar\n  package source is invalid\n',
        ),
        (
            'update foo:bar',
            '✗ foo:bar\n  package is not installed\n',
        ),
        (
            'remove foo:bar',
            '✗ foo:bar\n  package is not installed\n',
        ),
    ],
)
def test_invalid_source_lifecycle(run, args, output):
    assert run(args).output == output


@pytest.mark.parametrize(
    'args, cmp',
    [
        (
            'install curse:molinari',
            lambda v: v.startswith('✓ curse:molinari\n  installed'),
        ),
        (
            'install curse:20338',
            '✗ curse:20338\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowace.com/projects/molinari',
            '✗ curse:molinari\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.curseforge.com/wow/addons/molinari',
            '✗ curse:molinari\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.curseforge.com/wow/addons/molinari/download',
            '✗ curse:molinari\n  package already installed\n'.__eq__,
        ),
        (
            'remove curse:molinari',
            '✓ curse:molinari\n  removed\n'.__eq__,
        ),
    ],
)
def test_install_with_curse_alias(run, args, cmp):
    assert cmp(run(args).output)


@pytest.mark.parametrize(
    'args, cmp, run',
    [
        (
            'install tukui:-1',
            lambda v: v.startswith('✓ tukui:-1\n  installed'),
            Flavour.retail,
        ),
        (
            'install tukui:-1',
            '✗ tukui:-1\n  package does not exist\n'.__eq__,
            Flavour.classic,
        ),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package already installed\n'.__eq__,
            Flavour.retail,
        ),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package does not exist\n'.__eq__,
            Flavour.classic,
        ),
        (
            'install https://www.tukui.org/download.php?ui=tukui',
            '✗ tukui:tukui\n  package already installed\n'.__eq__,
            Flavour.retail,
        ),
        (
            'install https://www.tukui.org/download.php?ui=tukui',
            '✗ tukui:tukui\n  package does not exist\n'.__eq__,
            Flavour.classic,
        ),
        (
            'install https://www.tukui.org/addons.php?id=1',
            lambda v: v.startswith('✓ tukui:1\n  installed'),
            None,
        ),
        (
            'install tukui:1',
            '✗ tukui:1\n  package already installed\n'.__eq__,
            None,
        ),
        (
            'remove tukui:1',
            '✓ tukui:1\n  removed\n'.__eq__,
            None,
        ),
    ],
    indirect=('run',),
)
def test_install_with_tukui_alias(run, args, cmp):
    assert cmp(run(args).output)


@pytest.mark.parametrize(
    'args, cmp',
    [
        (
            'install wowi:13188',
            lambda v: v.startswith('✓ wowi:13188\n  installed'),
        ),
        (
            'install wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/landing.php?fileid=13188',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/fileinfo.php?id=13188',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/download13188-Molinari',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/info13188-Molinari.html',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        (
            'remove wowi:13188',
            '✓ wowi:13188\n  removed\n'.__eq__,
        ),
    ],
)
def test_install_with_wowi_alias(run, args, cmp):
    assert cmp(run(args).output)


@pytest.mark.parametrize(
    'args, cmp',
    [
        (
            'install github:AdiAddons/AdiButtonAuras',
            lambda v: v.startswith('✓ github:AdiAddons/AdiButtonAuras\n  installed'),
        ),
        (
            'install github:adiaddons/adibuttonauras',
            '✗ github:adiaddons/adibuttonauras\n  package already installed\n'.__eq__,
        ),
        (
            'install https://github.com/AdiAddons/AdiButtonAuras',
            '✗ github:AdiAddons/AdiButtonAuras\n  package already installed\n'.__eq__,
        ),
        (
            'install https://github.com/AdiAddons/AdiButtonAuras/releases',
            '✗ github:AdiAddons/AdiButtonAuras\n  package already installed\n'.__eq__,
        ),
        (
            'remove github:adiaddons/adibuttonauras',
            '✓ github:adiaddons/adibuttonauras\n  removed\n'.__eq__,
        ),
    ],
)
def test_install_with_github_alias(run, args, cmp):
    assert cmp(run(args).output)


@pytest.mark.parametrize(
    'args, output, run',
    [
        (
            'install curse:molinari',
            '✓ curse:molinari\n  installed 80300.66-Release\n',
            None,
        ),
        (
            'install --version foo curse:molinari',
            '✗ curse:molinari\n  package already installed\n',
            None,
        ),
        (
            'update curse:molinari',
            '✗ curse:molinari\n  package is up to date\n',
            None,
        ),
        (
            'remove curse:molinari',
            '✓ curse:molinari\n  removed\n',
            None,
        ),
        (
            'install --version foo curse:molinari',
            "✗ curse:molinari\n  no files compatible with retail using 'version' strategy\n",
            Flavour.retail,
        ),
        (
            'install --version foo curse:molinari',
            "✗ curse:molinari\n  no files compatible with classic using 'version' strategy\n",
            Flavour.classic,
        ),
        (
            'install --version 80000.57-Release curse:molinari',
            '✓ curse:molinari\n  installed 80000.57-Release\n',
            None,
        ),
        (
            'update curse:molinari',
            '✗ curse:molinari\n  package is up to date\n',
            None,
        ),
        (
            'remove curse:molinari',
            '✓ curse:molinari\n  removed\n',
            None,
        ),
    ],
    indirect=('run',),
)
def test_version_strategy_lifecycle(run, args, output):
    assert run(args).output == output


def test_install_sandwich(run):
    assert (
        run(
            'install'
            ' curse:molinari'
            ' -s latest curse:molinari'
            ' --version 80000.57-Release curse:molinari'
        ).output
        == '''\
✓ curse:molinari
  installed 80300.66-Release
✗ curse:molinari
  package folders conflict with installed package curse:molinari
✗ curse:molinari
  package folders conflict with installed package curse:molinari
'''
    )


def test_install_sandwich_defn_order_is_respected(run):
    assert (
        run(
            'install'
            ' --version 80000.57-Release curse:molinari'
            ' -s latest curse:molinari'
            ' curse:molinari'
        ).output
        == '''\
✓ curse:molinari
  installed 80000.57-Release
✗ curse:molinari
  package folders conflict with installed package curse:molinari
✗ curse:molinari
  package folders conflict with installed package curse:molinari
'''
    )


def test_install_sandwich_addon_argument_is_not_required(run):
    assert (
        run(
            'install'
            ' -s latest curse:molinari'
            ' --version 80000.57-Release curse:molinari'
        ).output
        == '''\
✓ curse:molinari
  installed 80300.66-Release
✗ curse:molinari
  package folders conflict with installed package curse:molinari
'''
    )


def test_addon_is_removed_even_if_folders_missing(cli_config, molinari_and_run):
    cli_config.addon_dir.joinpath('Molinari').rename(cli_config.addon_dir.joinpath('NotMolinari'))
    assert {cli_config.addon_dir.joinpath('NotMolinari')} == set(cli_config.addon_dir.iterdir())
    assert molinari_and_run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'


@pytest.mark.parametrize(
    'command, exit_code',
    [
        ('list mol', 0),
        ('info foo', 0),
        ('reveal mol', 0),
        ('reveal foo', 1),
    ],
)
def test_exit_codes_with_substr_match(monkeypatch, molinari_and_run, command, exit_code):
    monkeypatch.setattr('webbrowser.open', lambda v: ...)
    assert molinari_and_run(command).exit_code == exit_code


def test_can_list_with_substr_match(molinari_and_run):
    assert molinari_and_run('list mol').output == 'curse:molinari\n'
    assert molinari_and_run('list foo').output == ''
    assert molinari_and_run('list -f detailed mol').output.startswith('curse:molinari')
    (molinari,) = json.loads(molinari_and_run('list -f json mol').output)
    assert (molinari['source'], molinari['slug']) == ('curse', 'molinari')


def test_json_export_and_import(cli_config, molinari_and_run):
    export_json = cli_config.config_dir.parent / 'export.json'
    export_json.write_text(molinari_and_run('list -f json').output, encoding='utf-8')
    assert (
        molinari_and_run(f'install --import "{export_json}"').output
        == '✗ curse:molinari\n  package already installed\n'
    )



from __future__ import annotations

import asyncio
from collections import defaultdict
from contextlib import contextmanager
from functools import partial
from itertools import repeat
import os
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Optional as O,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from aiohttp import web
from aiohttp_rpc import JsonRpcMethod, WsJsonRpcServer, middlewares as rpc_middlewares
from aiohttp_rpc.errors import InvalidParams as InvalidParamsError, ServerError
from pydantic import BaseModel, ValidationError, validator

from . import exceptions as E
from .config import Config
from .manager import Manager, init_web_client
from .matchers import get_folders, match_dir_names, match_toc_ids, match_toc_names
from .models import Pkg, is_pkg
from .resolvers import Defn, PkgModel, Strategies
from .utils import Literal, get_version, is_outdated, run_in_thread as t, uniq

if TYPE_CHECKING:
    _T = TypeVar('_T')
    ManagerWorkQueueItem = Tuple[asyncio.Future[Any], str, O[Callable[..., Awaitable[Any]]]]


API_VERSION = 0


class _ConfigError(ServerError):
    code = -32001
    message = 'invalid configuration parameters'


@contextmanager
def _reraise_validation_error(error_class: Type[ServerError] = ServerError) -> Iterator[None]:
    try:
        yield
    except ValidationError as error:
        raise error_class(data=error.errors()) from error


class BaseParams(BaseModel):
    _method: ClassVar[str]

    async def respond(self, managers: ManagerWorkQueue) -> Any:
        raise NotImplementedError


class _ProfileParamMixin(BaseModel):
    profile: str


class _DefnParamMixin(BaseModel):
    defns: List[Defn]


class WriteConfigParams(BaseParams):
    values: Dict[str, Any]
    _method = 'config/write'

    @t
    def respond(self, managers: ManagerWorkQueue) -> Config:
        with _reraise_validation_error(_ConfigError):
            config = Config(**self.values).write()

        # Dispose of the ``Manager`` corresponding to the profile so that it is
        # re-loaded on next invocation of ``ManagerWorkQueue.run``
        managers.unload(config.profile)
        return config


class ReadConfigParams(_ProfileParamMixin, BaseParams):
    _method = 'config/read'

    @t
    def respond(self, managers: ManagerWorkQueue) -> Config:
        with _reraise_validation_error(_ConfigError):
            return Config.read(self.profile)


class DeleteConfigParams(_ProfileParamMixin, BaseParams):
    _method = 'config/delete'

    async def respond(self, managers: ManagerWorkQueue) -> None:
        async def delete_profile(manager: Manager):
            await t(manager.config.delete)()
            managers.unload(self.profile)

        await managers.run(self.profile, delete_profile)


class ListProfilesParams(BaseParams):
    _method = 'config/list'

    @t
    def respond(self, managers: ManagerWorkQueue) -> List[str]:
        return Config.list_profiles()


class Source(BaseModel):
    source: str
    name: str
    supported_strategies: List[Strategies]
    supports_rollback: bool

    @validator('supported_strategies')
    def _sort_strategies(cls, value: List[Strategies]) -> List[Strategies]:
        return sorted(value, key=list(Strategies).index)


class ListSourcesParams(_ProfileParamMixin, BaseParams):
    _method = 'sources/list'

    async def respond(self, managers: ManagerWorkQueue) -> List[Source]:
        manager = await managers.run(self.profile)
        return [
            Source(
                source=r.source,
                name=r.name,
                supported_strategies=r.strategies,
                supports_rollback=r.supports_rollback,
            )
            for r in manager.resolvers.values()
        ]


class ListResult(BaseModel):
    __root__: List[PkgModel]


class ListInstalledParams(_ProfileParamMixin, BaseParams):
    _method = 'list'

    async def respond(self, managers: ManagerWorkQueue) -> ListResult:
        installed_pkgs = await managers.run(
            self.profile, t(lambda m: m.database.query(Pkg).order_by(Pkg.name).all())
        )
        return ListResult.parse_obj(installed_pkgs)


class SuccessResult(BaseModel):
    status: Literal['success']
    addon: PkgModel


class ErrorResult(BaseModel):
    status: Literal['failure', 'error']
    message: str


class MultiResult(BaseModel):
    __root__: List[Union[SuccessResult, ErrorResult]]

    @validator('__root__', each_item=True, pre=True)
    def _classify_tuple(cls, value: Tuple[str, Any]) -> Union[SuccessResult, ErrorResult]:
        status, result = value
        if status == 'success':
            return SuccessResult(status=status, addon=result)
        else:
            return ErrorResult(status=status, message=result)


class SearchParams(_ProfileParamMixin, BaseParams):
    search_terms: str
    limit: int
    strategy: Strategies = Strategies.default
    _method = 'search'

    async def respond(self, managers: ManagerWorkQueue) -> MultiResult:
        results = await managers.run(
            self.profile,
            partial(
                Manager.search,
                search_terms=self.search_terms,
                limit=self.limit,
                strategy=self.strategy,
            ),
        )
        return MultiResult.parse_obj(list(zip(repeat('success'), results.values())))


class ResolveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    _method = 'resolve'

    async def respond(self, managers: ManagerWorkQueue) -> MultiResult:
        def extract_source(manager: Manager, defns: List[Defn]):
            for defn in defns:
                if defn.source == '*':
                    pair = manager.pair_uri(defn.alias)
                    if pair:
                        source, alias = pair
                        defn = defn.with_(source=source, alias=alias)
                yield defn

        results = await managers.run(
            self.profile,
            partial(
                Manager.resolve,
                defns=list(extract_source(await managers.run(self.profile), self.defns)),
            ),
        )
        return MultiResult.parse_obj(
            [('success', r) if is_pkg(r) else (r.kind, r.message) for r in results.values()]
        )


class InstallParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    replace: bool
    _method = 'install'

    async def respond(self, managers: ManagerWorkQueue) -> MultiResult:
        results = await managers.run(
            self.profile, partial(Manager.install, defns=self.defns, replace=self.replace)
        )
        return MultiResult.parse_obj(
            [
                (r.kind, r.pkg if isinstance(r, E.PkgInstalled) else r.message)
                for r in results.values()
            ]
        )


class UpdateParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    _method = 'update'

    async def respond(self, managers: ManagerWorkQueue) -> MultiResult:
        results = await managers.run(
            self.profile, partial(Manager.update, defns=self.defns, retain_strategy=True)
        )
        return MultiResult.parse_obj(
            [
                (r.kind, r.new_pkg if isinstance(r, E.PkgUpdated) else r.message)
                for r in results.values()
            ]
        )


class RemoveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    _method = 'remove'

    async def respond(self, managers: ManagerWorkQueue) -> MultiResult:
        results = await managers.run(self.profile, partial(Manager.remove, defns=self.defns))
        return MultiResult.parse_obj(
            [
                (r.kind, r.old_pkg if isinstance(r, E.PkgRemoved) else r.message)
                for r in results.values()
            ]
        )


class PinParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    _method = 'pin'

    async def respond(self, managers: ManagerWorkQueue) -> MultiResult:
        results = await managers.run(self.profile, partial(Manager.pin, defns=self.defns))
        return MultiResult.parse_obj(
            [
                (r.kind, r.pkg if isinstance(r, E.PkgInstalled) else r.message)
                for r in results.values()
            ]
        )


class AddonFolder(BaseModel):
    name: str
    version: str


class AddonMatch(BaseModel):
    folders: List[AddonFolder]
    matches: List[PkgModel]


class ReconcileResult(BaseModel):
    reconciled: List[AddonMatch]
    unreconciled: List[AddonMatch]

    @validator('reconciled', 'unreconciled', pre=True)
    def _transform_matches(cls, value: Any):
        return [
            {
                'folders': [AddonFolder(name=f.name, version=f.version) for f in s],
                'matches': m,
            }
            for s, m in value
        ]


_matchers = {
    'toc_ids': match_toc_ids,
    'dir_names': match_dir_names,
    'toc_names': match_toc_names,
}


class ReconcileParams(_ProfileParamMixin, BaseParams):
    matcher: Literal['toc_ids', 'dir_names', 'toc_names']
    _method = 'reconcile'

    async def respond(self, managers: ManagerWorkQueue) -> ReconcileResult:
        leftovers = await managers.run(self.profile, t(get_folders))
        match_groups = await managers.run(
            self.profile, partial(_matchers[self.matcher], leftovers=leftovers)
        )
        uniq_defns = uniq(d for _, b in match_groups for d in b)
        resolve_results = await managers.run(
            self.profile, partial(Manager.resolve, defns=uniq_defns)
        )
        reconciled = [
            (a, list(filter(is_pkg, (resolve_results[i] for i in d)))) for a, d in match_groups
        ]
        unreconciled = [
            ([l], []) for l in sorted(leftovers - frozenset(i for a, _ in match_groups for i in a))
        ]
        return ReconcileResult(reconciled=reconciled, unreconciled=unreconciled)


class GetVersionResult(BaseModel):
    installed_version: str
    new_version: O[str]


class GetVersionParams(BaseParams):
    _method = 'meta/get_version'

    async def respond(self, managers: ManagerWorkQueue) -> GetVersionResult:
        outdated, new_version = await t(is_outdated)()
        return GetVersionResult(
            installed_version=get_version(), new_version=new_version if outdated else None
        )


class ManagerWorkQueue:
    def __init__(self) -> None:
        asyncio.get_running_loop()  # Sanity check
        self._queue: asyncio.Queue[ManagerWorkQueueItem] = asyncio.Queue()
        self._managers: Dict[str, Manager] = {}
        self._locks: DefaultDict[str, asyncio.Lock] = defaultdict(asyncio.Lock)
        self._web_client = init_web_client()

    async def _jumpstart(self, profile: str) -> Manager:
        try:
            manager = self._managers[profile]
        except KeyError:
            async with self._locks['load manager']:
                with _reraise_validation_error(_ConfigError):
                    config = await t(Config.read)(profile)
                self._managers[profile] = manager = await t(Manager.from_config)(config)
            manager.web_client = self._web_client
            manager.locks = self._locks
        return manager

    async def listen(self) -> None:
        while True:
            (future, profile, coro_fn) = item = await self._queue.get()
            try:
                manager = await self._jumpstart(profile)
            except BaseException as error:
                future.set_exception(error)
            else:
                if coro_fn:

                    async def schedule(item: ManagerWorkQueueItem, manager: Manager):
                        future, _, coro_fn = item
                        try:
                            result = await coro_fn(manager)  # type: ignore
                        except BaseException as error:
                            future.set_exception(error)
                        else:
                            future.set_result(result)

                    asyncio.create_task(schedule(item, manager))
                else:
                    future.set_result(manager)

            self._queue.task_done()

    async def cleanup(self) -> None:
        await self._web_client.close()

    @overload
    async def run(self, profile: str, coro_fn: None = ...) -> Manager:
        ...

    @overload
    async def run(self, profile: str, coro_fn: Callable[..., Awaitable[_T]] = ...) -> _T:
        ...

    async def run(
        self, profile: str, coro_fn: O[Callable[..., Awaitable[_T]]] = None
    ) -> Union[Manager, _T]:
        future: asyncio.Future[Any] = asyncio.Future()
        self._queue.put_nowait((future, profile, coro_fn))
        return await asyncio.wait_for(future, None)

    def unload(self, profile: str) -> None:
        self._managers.pop(profile, None)


def _prepare_response(param_class: Type[BaseParams], managers: ManagerWorkQueue) -> JsonRpcMethod:
    async def respond(**kwargs: Any) -> BaseModel:
        with _reraise_validation_error(InvalidParamsError):
            params = param_class.parse_obj(kwargs)
        return await params.respond(managers)

    return JsonRpcMethod('', respond, custom_name=param_class._method)


def serialise_response(value: Dict[str, Any]) -> str:
    return BaseModel.construct(**value).json()


async def create_app() -> web.Application:
    managers = ManagerWorkQueue()
    rpc_server = WsJsonRpcServer(
        json_serialize=serialise_response,
        middlewares=rpc_middlewares.DEFAULT_MIDDLEWARES,
    )
    rpc_server.add_methods(
        map(
            _prepare_response,
            (
                WriteConfigParams,
                ReadConfigParams,
                DeleteConfigParams,
                ListProfilesParams,
                ListSourcesParams,
                ListInstalledParams,
                SearchParams,
                ResolveParams,
                InstallParams,
                UpdateParams,
                RemoveParams,
                PinParams,
                ReconcileParams,
                GetVersionParams,
            ),
            repeat(managers),
        )
    )
    app = web.Application()
    app.router.add_routes([web.get(f'/v{API_VERSION}', rpc_server.handle_http_request)])

    async def on_shutdown(app: web.Application):
        listen.cancel()
        await managers.cleanup()

    listen = asyncio.create_task(managers.listen())
    app.on_shutdown.append(on_shutdown)
    return app


async def listen() -> None:
    "Fire up the server."
    loop = asyncio.get_running_loop()
    app_runner = web.AppRunner(await create_app())
    await app_runner.setup()
    # Placate the type checker - server is created in ``app_runner.setup``
    assert app_runner.server
    # By omitting the port ``loop.create_server`` will find a random available
    # port to bind to (equivalent to creating a socket on port 0)
    server = await loop.create_server(app_runner.server, '127.0.0.1')
    assert server.sockets
    (host, port) = server.sockets[0].getsockname()
    # We're writing the address to fd 3 just in case something seeps into
    # stdout or stderr.
    message = f'ws://{host}:{port}/'.encode()
    for fd in (1, 3):
        os.write(fd, message)
    try:
        # ``server_forever`` cleans up after the server when it's interrupted
        await server.serve_forever()
    finally:
        await app_runner.cleanup()
